#![deny(clippy::pedantic)]

use std::{
    cmp::Ordering,
    ffi::OsStr,
    fs::OpenOptions,
    io::Read,
    path::PathBuf,
    process::{exit, Stdio},
    str::FromStr,
};
use tokio::io::AsyncBufReadExt;
use tokio::process::Command;
use tokio::{io::BufReader, join};

use anyhow::{anyhow, bail, Context};
use domain::{
    base::{Dname, RelativeDname},
    resolv::StubResolver,
};
use lettre::{
    address::Envelope, transport::smtp::authentication::Credentials, Address, SmtpTransport,
    Transport,
};
use mailparse::{parse_mail, MailAddr, MailHeaderMap, ParsedMail};

#[derive(serde::Deserialize, Debug)]
struct Config {
    account: String,
    passcmd: Vec<String>,
}

impl Config {
    fn load() -> anyhow::Result<Config> {
        let config_home = if let Some(s) = std::env::var_os("XDG_CONFIG_HOME") {
            PathBuf::from(s).join("xendmail")
        } else if let Some(home) = std::env::var_os("HOME") {
            PathBuf::from(home).join(".config/xendmail")
        } else if let Ok(cwd) = std::env::current_dir() {
            cwd
        } else {
            bail!("No viable candidate for configuration directory");
        };

        let mut config_file = OpenOptions::new()
            .read(true)
            .write(false)
            .create(false)
            .open(config_home.join("default.toml"))
            .context("failed to load configuration file")?;

        // toml crate won't allow reading from a file.
        // See: https://github.com/toml-rs/toml/pull/349
        let mut config = Vec::new();
        config_file.read_to_end(&mut config)?;

        Ok(toml::de::from_str(std::str::from_utf8(&config)?)?)
    }

    async fn password(&self) -> anyhow::Result<String> {
        let mut parts = self.passcmd.iter();
        let mut child = Command::new(
            parts
                .next()
                .context("passcmd must contain at least one value")?,
        )
        .args(parts.collect::<Vec<_>>())
        .stdout(Stdio::piped())
        .spawn()?;

        let mut line = String::new();
        BufReader::new(child.stdout.take().expect("child has a piped stdout"))
            .read_line(&mut line)
            .await?;
        let status = child.wait().await?;
        if !status.success() {
            bail!("Password command exited with status code: {}.", status);
        }
        if line.ends_with('\n') {
            line.pop();
        }
        Ok(line)
    }

    async fn server_and_port(&self) -> anyhow::Result<(String, u16)> {
        let email = Address::from_str(&self.account).context("error parsing email address")?;
        let host = email.domain();

        let dname = Dname::bytes_from_str(host)?;
        let candidates = resolve_server(&dname, 587)
            .await
            .context("Failed to resolve SRV records")?;

        let (host, port) = candidates
            .first()
            .context("No results when querying for submission server")?;
        Ok((host.to_string(), *port))
    }
}

/// Resolves a submission server for a given domain.
///
/// See: <https://www.rfc-editor.org/rfc/rfc6186>
///
/// # Errors
///
/// If the DNS query fails or returns invalid data.
pub async fn resolve_server<T: std::convert::AsRef<[u8]>>(
    domain: &Dname<T>,
    port: u16,
) -> anyhow::Result<Vec<(String, u16)>> {
    let service = RelativeDname::from_slice(b"\x0b_submission\x04_tcp")
        .expect("submission relative domain is valid");
    let response = StubResolver::new()
        .lookup_srv(service, domain, port)
        .await
        .map_err(|e| anyhow::anyhow!("Error querying SRV records: {:?}", e))?;

    let mut srvs: Vec<_> = match response {
        Some(s) => s.into_srvs().collect(),
        None => return Ok(vec![]),
    };

    if let Some(srv) = srvs.first() {
        if srv.target().as_ref() == b"." {
            bail!("Submission service decidedly not available");
        }
    }

    // A client MUST attempt to contact the target host with the lowest-numbered priority it can reach[...]
    // [...] Larger weights SHOULD be given a proportionately higher probability of being selected. [...]
    srvs.sort_unstable_by(|s1, s2| {
        match s1.priority().cmp(&s2.priority()) {
            Ordering::Less => Ordering::Less,
            Ordering::Equal => s2.weight().cmp(&s1.weight()), // Hint: in reverse order!
            Ordering::Greater => Ordering::Greater,
        }
    });

    Ok(srvs
        .iter()
        .map(|s| (s.target().to_string(), s.port()))
        .collect())
}

fn extract_from(message: &ParsedMail) -> anyhow::Result<Address> {
    message
        .get_headers()
        .get_first_header("From")
        .map(mailparse::addrparse_header)
        .transpose()?
        .and_then(|addr_list| addr_list.into_inner().into_iter().next())
        .ok_or(anyhow!("Missing 'From' header"))
        .map(|addr| match addr {
            MailAddr::Single(single) => {
                Address::from_str(&single.addr).context("Error parsing address")
            }
            MailAddr::Group(_) => Err(anyhow::anyhow!(
                "The 'From' header cannot contain more than one address"
            )),
        })?
}

fn extract_recipients(message: &ParsedMail) -> Vec<Address> {
    let headers = message.get_headers();
    let recipients = message
        .headers
        .get_all_headers("To")
        .into_iter()
        .chain(headers.get_all_headers("Cc").into_iter())
        .chain(headers.get_all_headers("Bcc").into_iter())
        .flat_map(mailparse::addrparse_header)
        .flat_map(|addr_list| {
            addr_list
                .into_inner()
                .into_iter()
                .map(|addr| match addr {
                    MailAddr::Group(group) => group.addrs,
                    MailAddr::Single(single) => vec![single],
                })
                .collect::<Vec<_>>()
        })
        .flatten()
        .flat_map(|single| single.addr.parse())
        .collect::<Vec<_>>();

    recipients
}

/// An email enqueuer for remote SMTP servers.
pub(crate) struct Args {
    /// Recipients of the email message.
    ///
    /// If no recipients are provided, the `To`, `Cc` and `Bcc` fields are used.
    to: Vec<Address>,
}

impl Args {
    fn add_to(&mut self, raw: &OsStr) -> anyhow::Result<()> {
        self.to.push(
            raw.to_str()
                .with_context(|| format!("Email is not valid UTF-8: {}", raw.to_string_lossy()))?
                .parse()?,
        );
        Ok(())
    }
    fn parse(raw: impl IntoIterator<Item = impl Into<std::ffi::OsString>>) -> anyhow::Result<Self> {
        let mut args = Args { to: Vec::new() };

        let raw = clap_lex::RawArgs::new(raw);
        let mut cursor = raw.cursor();
        raw.next(&mut cursor); // Skip the program's name.
        while let Some(arg) = raw.next(&mut cursor) {
            if arg.is_escape() {
                let remaining = raw.remaining(&mut cursor);
                args.to.reserve({
                    let (min, max) = remaining.size_hint();
                    max.unwrap_or(min)
                });
                for r in remaining {
                    args.add_to(r)?;
                }
            } else if arg.is_stdio() {
                bail!("Invalid argument '-'");
            } else if let Some((_long, _value)) = arg.to_long() {
                bail!("Invalid argument '{}'", arg.display());
            } else if let Some(mut shorts) = arg.to_short() {
                while let Some(short) = shorts.next_flag() {
                    match short {
                        Ok('A' | 'B' | 'b' | 'E' | 'e' | 'i' | 'L' | 'm' | 'o' | 'p' | 'x') => {
                            // No-op (sendmail compat)
                        }
                        Ok(c) => {
                            bail!("Unexpected flag: -{c}");
                        }
                        Err(e) => {
                            bail!("Invalid flag: -{}", e.to_string_lossy());
                        }
                    }
                }
            } else {
                args.add_to(arg.to_value_os())?;
            }
        }

        Ok(args)
    }
}

/// Docs: <https://man.openbsd.org/sendmail>
fn main() -> anyhow::Result<()> {
    let args = Args::parse(std::env::args_os())?;
    let mut stdin = std::io::stdin();

    let mut raw = Vec::new();
    stdin.read_to_end(&mut raw)?;
    let parsed = parse_mail(&raw)?;

    let to = if args.to.is_empty() {
        extract_recipients(&parsed)
    } else {
        args.to
    };

    let from = extract_from(&parsed)?;
    let envelope = Envelope::new(Some(from), to)?;
    // TODO: Remove Bcc (need to reconstruct the email for that)

    let config = Config::load()?;
    let (password_result, sp_result) = tokio::runtime::Builder::new_current_thread()
        .enable_all()
        .build()?
        .block_on(async { join!(config.password(), config.server_and_port()) });
    let password = password_result.context("Failed to determine password")?;
    let (server, port) = sp_result.context("Failed to resolve SRV records")?;

    let sender = SmtpTransport::starttls_relay(&server)?
        .port(port)
        .credentials(Credentials::new(config.account, password))
        .build();

    // The `Message` type requires `builder`.
    let response = sender.send_raw(&envelope, &raw)?;
    if !response.is_positive() {
        eprintln!("Server returned an error response: {:?}", &response);
        exit(1);
    }

    Ok(())
}
